local M = {}
M.type = "auxiliary"
M.relevantDevice = nil
M.defaultOrder = 1000

local constants = {
    -- Conversion rates
    mpsToMph = 2.23694,
    hpToKw = 0.7457,
    torqueToHpCoef = 7120.54016,
    kWhToJ = 3600000,
    sToH = 1.0 / 3600,
    -- Sounds
    inverterBaseVolume = 0.85,
    inverterMaxVolumeKw = 50,
    inverterPitchMinSpeed = 0.25,
    inverterPitchMaxSpeed = 2,
    motorBaseVolume = 0.45,
    motorVolumeMinSpeed = 0.2,
    motorVolumeMaxSpeed = 50,
    motorPitchMinSpeed = 0,
    motorPitchMaxSpeed = 95,
    motorMinPitch = 0.3,
    motorMaxPitch = 3.4
}

local min = math.min
local max = math.max
local abs = math.abs
local mathutil = require("mathutil")
local debugutil = require("debugutil")("hybrid", false)
local maprange = mathutil.maprange
local clamp = mathutil.clamp
local blend = mathutil.blend

local engine = nil
local gearbox = nil
local battery = nil
local esc = nil

local sounds = {
    inverterWhine = nil,
    motorRamp = nil
}

-- Calculated/initial values
local initialWheelBrakeSplit = {}

-- Quick access upvalues
local engineOn = false
local watertemp = 0
local speedMph = 0
local batteryLevel = 1
local totalPowerOutKw = 0
local throttle = 0
local throttleAggressiveness = 0
local gear = "P"

-- Other state variables
local engineDemandReason = "none"
local state = "init"
local nextState = nil
local timeSinceEngineStart = 0
local timeSinceAppliedBrakes = 0
local startingTime = 0
local isInChargeup = false
local lastThrottle = 0
local chargeThrottle = 0
local disableRegen = false
local timeSinceAbsHit = 0
local timeSinceInit = 0

local function isEngineDemandedOn()
    local maxEvOnlyThrottle

    if gear == "P" then
        maxEvOnlyThrottle = 0.02
    else
        maxEvOnlyThrottle = M.maxEvOnlyThrottle * maprange(speedMph, M.maxEvOnlySpeed / 2, M.maxEvOnlySpeed, 1.0, 0.5)
    end

    if not M.enableAutoStop then
        engineDemandReason = "auto stop disabled"
        return true
    end
    if watertemp < 57 then
        engineDemandReason = "temp"
        return true
    end -- engine must be warmed up
    if speedMph > M.maxEvOnlySpeed then
        engineDemandReason = "speed"
        return true
    end -- too fast
    if batteryLevel < M.minBatteryPowerLevel or isInChargeup then
        engineDemandReason = "battery level"
        isInChargeup = true
        return true
    end -- battery too low
    if timeSinceEngineStart < 4 then
        --engineDemandReason = "time since start"
        return true
    end -- engine has to run for at least 4 seconds
    if throttle > maxEvOnlyThrottle or throttleAggressiveness > 8 then
        engineDemandReason = "throttle"
        return true
    end
    if engineOn and (input.brake < M.autoStopBrakeThreshold or timeSinceAppliedBrakes < M.autoStopBrakeTime) then
        -- this check only applies when not in P
        if gear ~= "P" then
            engineDemandReason = "brakes"
            return true
        end
    end

    -- Easy checks are done; dynamic checks now

    local curPower = gearbox.outputKw

    if engineOn then
        if curPower > 2 then
            engineDemandReason = "accelerating"
            return true
        end -- once engine is on, must coast again to shut off
    else
        -- This is the percentage of the max EV-only kW output that's available to us,
        -- based on battery level
        local batteryWorkingRatio = maprange(batteryLevel, M.minBatteryPowerLevel, M.maxBatteryPowerLevel)
        local availableEvOnlyPower = batteryWorkingRatio * M.maxEvOnlyPower
        
        debugutil.dText(40, "bWR: "..batteryWorkingRatio)
        debugutil.dText(41, "aEOP: "..availableEvOnlyPower)
        debugutil.dText(42, "thp: "..(throttle / M.maxEvOnlyThrottle))

        if curPower > availableEvOnlyPower or throttle > batteryWorkingRatio * M.evPercentOfTotalPower then
            engineDemandReason = "accelerating"
            return true
        end
    end

    engineDemandReason = "none"
    return false -- by default, the engine is off
end

local function updateValues(dt)
    if not gearbox or not engine or not battery then
        return
    end

    engineOn = not (engine.isStalled or engine.ignitionCoef < 0.01)
    watertemp = electrics.values.watertemp
    speedMph = electrics.values.wheelspeed * constants.mpsToMph
    batteryLevel = battery.remainingRatio
    throttle = electrics.values.throttleOverride or input.throttle
    throttleAggressiveness = (throttle - lastThrottle) / dt
    gear = electrics.values.gear

    if engineOn then
        timeSinceEngineStart = timeSinceEngineStart + dt
    end
    
    if M.autoStopBrakeThreshold > 0 and input.brake >= M.autoStopBrakeThreshold and speedMph <= M.maxEvOnlySpeed then
        timeSinceAppliedBrakes = timeSinceAppliedBrakes + dt
    else
        timeSinceAppliedBrakes = 0
    end

    if isInChargeup and batteryLevel >= blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 0.5) then
        isInChargeup = false
    end
end

local function updateState(dt)
    local engineDemandedOn = isEngineDemandedOn()

    debugutil.dText(0, "Engine demanded on: " .. tostring(engineDemandedOn))
    debugutil.dText(1, "Reason: " .. engineDemandReason)
    debugutil.dText(2, "State: " .. state)
    
    -- keep things like daytime-running-lights on
    electrics.values.running = 1

    if state == "init" then
        timeSinceInit = timeSinceInit + dt
        
        if timeSinceInit >= 0.5 then
            if engineOn then
                nextState = "running"
            else
                nextState = "stopped"
            end
        end
    elseif state == "evmode" then
        gearbox:setInputEngaged(false)
        engine.ignitionCoef = 0 -- Kill engine
        engine.isDisabled = true -- Prevent the user from manually restarting the engine
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)

        if electrics.values.evMode ~= 1 then
            if M.forceEvMode then
                electrics.values.evMode = 1
            else
                nextState = "init"
            end
        end
    elseif state == "running" then
        local canEngageInGear = gear == "L" or gear == "D"
        local engineInRpmBounds = electrics.values.rpm >= engine.idleRPM
        local canEngineEngage = speedMph >= 2.5 or throttle > 0
        
        gearbox:setInputEngaged(canEngageInGear and engineInRpmBounds and canEngineEngage)

        if electrics.values.evMode == 1 then
            nextState = "evmode"
        elseif electrics.values.chargeMode == 1 and gear == "P" then
            chargeThrottle = 0
            nextState = "chargemode"
        elseif not engineDemandedOn then
            if gear ~= "N" then -- engine keeps state in N
                gearbox:setInputEngaged(false)
                engine.ignitionCoef = 0 -- Kill engine
                engine.isDisabled = true -- Prevent the user from manually restarting the engine
                electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)

                nextState = "stopped"
            end
        end
    elseif state == "chargemode" then
        local inverseRelativePowerLevel = maprange(batteryLevel, 0.9, 1.0, 1.0, 0.0)
        local curGenOutput = -min(totalPowerOutKw, 0)
        local targGenOutput = inverseRelativePowerLevel * M.maxFastChargePower
        local error = (targGenOutput - curGenOutput) / targGenOutput

        chargeThrottle = clamp(chargeThrottle + (error * dt * M.chargeThrottleIFactor), 0, 1)

        if gear ~= "P" or electrics.values.evMode == 1 or batteryLevel >= 0.95 then
            electrics.values.chargeMode = 0
        end

        if electrics.values.chargeMode ~= 1 then
            nextState = "init"
        end
    elseif state == "stopped" then
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        electrics.values.iceThrottle = 0
        gearbox:setInputEngaged(false)

        if electrics.values.evMode == 1 then
            nextState = "evmode"
        elseif gear ~= "N" and (engineDemandedOn or electrics.values.chargeMode == 1) then
            engine.isDisabled = false -- allow engine to restart

            if electrics.values.rpm < engine.idleRPM then
                engine:activateStarter() -- start engine
                nextState = "startup"
                timeSinceEngineStart = 0
                startingTime = 0
            else
                engine.ignitionCoef = 1 -- engine already spinning, engage ignition
                nextState = "running"
            end
        end
    elseif state == "startup" then
        electrics.values.iceThrottle = 0
        startingTime = startingTime + dt

        if engineOn then
            if timeSinceEngineStart > 0.15 then
                nextState = "running"
            end
        else
            if startingTime > 3 then
                -- Something's not right; retry at init

                nextState = "init"
            end
        end
    end
end

local function updateBattery(dt)
    totalPowerOutKw = M.idlePowerDraw

    local motorKw = gearbox.outputKw * (gear == "R" and -1 or 1)

    if motorKw > 0 then
        totalPowerOutKw = totalPowerOutKw + motorKw
    else
        totalPowerOutKw = totalPowerOutKw + (motorKw * M.regenEfficiency)
    end

    if engineOn then
        local requestedGenTorque = 0
        local inverseRelativePowerLevel = 0

        if state == "chargemode" then
            inverseRelativePowerLevel = maprange(batteryLevel, 0.9, 1.0, 1.0, 0.0)
        else
            local minBat = blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 1 / 3)
            local maxBat = M.maxBatteryPowerLevel

            inverseRelativePowerLevel = maprange(batteryLevel, minBat, maxBat, 1.0, 0.0)
        end
        
        local maxGenCoef = 0
        
        if state == "chargemode" then
            maxGenCoef = maprange(electrics.values.rpm, engine.idleRPM, M.engineMaxGeneratorRpm, 0, 1)
        else
            maxGenCoef = 1 - throttle
        end

        local rpmCoef = maprange(electrics.values.rpm, 0, M.engineMaxGeneratorRpm * 0.9)
        local genCoef = maprange(rpmCoef, 0, inverseRelativePowerLevel, inverseRelativePowerLevel, maxGenCoef)

        requestedGenTorque = M.maxEngineGeneratorTorque * genCoef
        gearbox:setEngineGenerationTorque(requestedGenTorque)
        debugutil.dText(10, "iRPL: " .. tostring(inverseRelativePowerLevel) .. "; rEG: " .. tostring(requestedGenTorque))
        totalPowerOutKw = totalPowerOutKw - gearbox.generatorKw
        
        electrics.values.lockupClutchRatio = maprange(electrics.values.rpm, engine.idleRPM, M.engineMaxGeneratorRpm, 0, genCoef)
    end

    local totalPowerOutKj = totalPowerOutKw * dt

    debugutil.dText(3, "Battery power (kW): " .. totalPowerOutKw)
    debugutil.dText(4, "Generator power (kW): " .. gearbox.generatorKw)

    battery.storedEnergy = mathutil.clamp(battery.storedEnergy - totalPowerOutKj * 1000, 0, battery.initialStoredEnergy)
end

local function updateSounds()
    -- Update sounds
    local inverterPitchCoef = maprange(speedMph, constants.inverterPitchMinSpeed, constants.inverterPitchMaxSpeed)
    local inverterEffort = math.abs(gearbox.outputKw) / constants.inverterMaxVolumeKw
    local inverterVolume = constants.inverterBaseVolume * mathutil.clamp(inverterEffort, 0.0, 1.0)
    local inverterPitch = mathutil.clamp(inverterPitchCoef, 0.4, 1.0)

    obj:setVolumePitch(sounds.inverterWhine, inverterVolume, inverterPitch)

    local motorPitch =
        maprange(
        speedMph,
        constants.motorPitchMinSpeed,
        constants.motorPitchMaxSpeed,
        constants.motorMinPitch,
        constants.motorMaxPitch,
        false
    )
    local motorVolume =
        maprange(
        speedMph,
        constants.motorVolumeMinSpeed,
        constants.motorVolumeMaxSpeed,
        0,
        constants.motorBaseVolume,
        false
    )

    obj:setVolumePitch(sounds.motorRamp, motorVolume, motorPitch)
end

local function updateGUI(dt)
    local str = {
        soc = batteryLevel,
        kw = totalPowerOutKw,
        maxKw = gearbox and gearbox.maxMotorPower or 0
    }

    gui.send("hybrid", str)
end

local function updateGFX(dt)
    updateValues(dt)

    if gearbox and engine and battery then
        updateState(dt)
        updateBattery(dt)

        local evThrottle = 0.0
        local iceThrottle = 0.0
        local actualMotorPercentMaxTorque = 0.0
        local evThrottleFactor = nil
        local creep = maprange(speedMph, 0, 3.5, M.evCreepThrottle, 0)
        local gearboxMaxAdjustedMotorTorque = gearbox:getMaxMotorTorque() * gearbox.motorGearRatio

        if input.brake > 0 then
            -- Launch control
            if throttle > 0 then
                iceThrottle = maprange(throttle, 0, 0.2, 0, 0.4, true)
            end
            
            input.throttle = 0
            throttle = 0
            lastThrottle = 0
            throttleAggressiveness = 0
            
            if gear == "N" or gear == "P" then
                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Reset brake split coef
                        wheel.brakeInputSplit = initialWheelBrakeSplit[i][1]
                        wheel.brakeSplitCoef = initialWheelBrakeSplit[i][2]
                    end
                end
            else
                local maxBrakeTorque = 0.0
                local absActive = false

                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Find total braking torque
                        if wheel.brakeTorque > maxBrakeTorque then
                            maxBrakeTorque = wheel.brakeTorque
                        end

                        -- Determine if ABS active
                        if wheel.absActive then
                            absActive = true
                        end
                    end
                end

                if disableRegen then
                    timeSinceAbsHit = timeSinceAbsHit + dt

                    if timeSinceAbsHit > 1.0 then
                        disableRegen = false
                        timeSinceAbsHit = 0
                    end
                else
                    if absActive then
                        timeSinceAbsHit = timeSinceAbsHit + dt

                        if timeSinceAbsHit > 0.5 then
                            disableRegen = true
                            timeSinceAbsHit = 0
                        end
                    else
                        timeSinceAbsHit = 0
                    end
                end

                local speedSign = mathutil.sign(electrics.values.avgWheelAV)
                local regenSign = gear == "R" and -1 or 1
                local maxRegen =
                    disableRegen and 0 or maprange(speedMph, M.minRegenFadeSpeed, M.maxRegenFadeSpeed, 0, 1, true)
                local throttleOffset = max(maxRegen * 0.25, 0.01)
                local negThrottle = gear == "L" and maprange(throttle, throttleOffset, 0.0, 0.0, 1.0, true) or 0
                local regen =
                    maprange(input.brake + negThrottle, 0.0, M.regenPercentOfBrakingForce) * maxRegen * speedSign *
                    regenSign

                evThrottle = -clamp(regen, 0, maxRegen) * M.maxRegenCoef
                evThrottleFactor = 100 -- faster for brake application

                local actualRegen =
                    clamp(-gearbox.motorOutputTorque / (gearboxMaxAdjustedMotorTorque * M.maxRegenCoef), 0, 1)
                local brakesCoveredByRegen = actualRegen * M.regenPercentOfBrakingForce
                local neededFrictionBrake = 1 - brakesCoveredByRegen
                
                input.brake = neededFrictionBrake

                for i = 0, wheels.wheelCount - 1, 1 do
                    local wd = wheels.wheels[i]

                    if wd.isPropulsed then
                        wd.brakeInputSplit = 0
                        wd.brakeSplitCoef = initialWheelBrakeSplit[i][2] * max(neededFrictionBrake - M.regenPercentOfBrakingForce, 0)
                    else
                        wd.brakeInputSplit = neededFrictionBrake
                        wd.brakeSplitCoef = initialWheelBrakeSplit[i][2]
                    end
                end

                debugutil.dText(11, "Max Regen: " .. maxRegen)
                debugutil.dText(12, "Regen/maxBrake: " .. brakesCoveredByRegen)
                debugutil.dText(13, "Regen %: " .. tostring(regen * 100))
                debugutil.dText(14, "Friction %: " .. tostring(neededFrictionBrake * 100))
            end
        else
            disableRegen = false

            local throttleFactor = electrics.values.throttleFactor or 1.0
            local availableEvThrottle = 0

            if state == "evmode" then
                availableEvThrottle = maprange(batteryLevel, 0.15, M.minBatteryPowerLevel)
            else
                availableEvThrottle = maprange(batteryLevel, M.minBatteryPowerLevel, M.maxBatteryPowerLevel)
            end

            actualMotorPercentMaxTorque = 1.0 - maprange(throttle, availableEvThrottle, 1.01, 0.0, 1.0, true)
            
            local iceThrottleSource = 0

            if gear == "L" then
                -- Regenerative braking when letting off accelerator

                local speedSign = mathutil.sign(electrics.values.avgWheelAV)
                local regenSign = gear == "R" and -1 or 1
                local maxRegen = maprange(speedMph, M.minRegenFadeSpeed, M.maxRegenFadeSpeed)
                local throttleOffset = max(maxRegen * 0.25, 0.01)
                local posThrottle = maprange(throttle, throttleOffset, 1.0, 0.0, 1.0, true)
                local negThrottle = maprange(throttle, throttleOffset, 0.0, 0.0, 1.0, true)
                local posEvThrottle = maprange(posThrottle, 0, M.evPercentOfTotalPower * max(availableEvThrottle, 0.01)) * availableEvThrottle
                local negEvThrottle = negThrottle * M.maxRegenCoef * maxRegen * regenSign * speedSign

                evThrottle = posEvThrottle - negEvThrottle
                iceThrottleSource = posThrottle
            else
                local minEvThrottle = maprange(speedMph, M.minRegenFadeSpeed, M.maxRegenFadeSpeed, 0.0, M.evCoastThrottle)
                
                evThrottle = maprange(throttle, 0, M.evPercentOfTotalPower * max(availableEvThrottle, 0.01), minEvThrottle, availableEvThrottle)
                iceThrottleSource = throttle
            end
            
            local iceThrottleThreshold = availableEvThrottle * M.evPercentOfTotalPower
            
            evThrottleFactor =
                maprange(
                throttleAggressiveness,
                M.minThrottleAggressivenessRamp,
                M.maxThrottleAggressivenessRamp,
                M.minThrottleAggressivenessFactor,
                M.maxThrottleAggressivenessFactor
            )
            iceThrottle = maprange(iceThrottleSource, iceThrottleThreshold, 1, 0, 1)
        end

        if input.brake < 0.25 and gear ~= "P" and gear ~= "N" then
            evThrottle = evThrottle + creep
        end

        gearbox:setElectricTorqueDemand(evThrottle, evThrottleFactor)
        debugutil.dText(8, "iT: " .. iceThrottle)
        debugutil.dText(9, "eT: " .. evThrottle .. ", mpmt: "..actualMotorPercentMaxTorque)

        local motorPowerPercent = gearbox.outputKw / gearbox.maxMotorPower
        
        if isInChargeup then
            iceThrottle = max(iceThrottle, M.engineGeneratorMaxMinThrottle)
        else
            local maxMinThrottle = maprange(batteryLevel, M.minBatteryPowerLevel, M.maxBatteryPowerLevel, M.engineGeneratorMaxMinThrottle, 0)
            
            iceThrottle = max(iceThrottle, maxMinThrottle)
        end

        if state == "running" then
            if gear == "R" then
                electrics.values.iceThrottle = min(iceThrottle, 0.05)
            else
                electrics.values.iceThrottle = iceThrottle
            end
        elseif state == "chargemode" then
            electrics.values.iceThrottle = chargeThrottle
        end

        if not gearbox.inputEngaged then
            -- In order to prevent over-reving when the ICE engages,
            -- we have to predict a close gear ratio range and keep
            -- the transmission in that range while the ICE is disengaged
            local estGearRatio =
                maprange(speedMph, M.minGearRatioSpeed, M.maxGearRatioSpeed, gearbox.maxGearRatio, gearbox.minGearRatio)

            gearbox:setGearRatio(estGearRatio)
        end
    end

    if battery then
        debugutil.dText(5, "Battery %: " .. tostring(battery.remainingRatio * 100))
    end
    debugutil.dText(6, "Throttle: " .. (throttle or 0) .. " (ICE: " .. (electrics.values.iceThrottle or 0) .. ")")
    debugutil.dText(7, "tA: " .. throttleAggressiveness)

    if nextState then
        state = nextState
    end

    lastThrottle = throttle

    debugutil.dText(17, "engTor: " .. engine.outputTorque1)
    debugutil.dText(18, "chgThr: " .. chargeThrottle)
    debugutil.dText(19, "kW: " .. gearbox.outputKw)
    debugutil.dText(20, "rpm: " .. gearbox.outputAV1 * 9.549)
    debugutil.dText(21, "maxT: " .. gearbox:getMaxMotorTorque() * gearbox.motorGearRatio)
    debugutil.dText(22, "mT: " .. gearbox.motorOutputTorque)
    debugutil.dText(23, "GR: " .. gearbox.gearRatio)

    updateGUI(dt)
    updateSounds()
end

local function init(jbeamData)
    M.jbeamData = jbeamData
    M.enableAutoStop = jbeamData.enableAutoStop ~= false
    M.maxEvOnlyPower = jbeamData.maxEvOnlyPower or 85 -- kW
    M.maxEvOnlySpeed = jbeamData.maxEvOnlySpeed or 45 -- MPH
    M.maxEvOnlyThrottle = jbeamData.maxEvOnlyThrottle or 0.45 -- ratio
    M.minGearRatioSpeed = jbeamData.minGearRatioSpeed or 10
    M.maxGearRatioSpeed = jbeamData.maxGearRatioSpeed or 90
    M.minBatteryPowerLevel = 0.25 -- ratio; low end of max-kW slope
    M.maxBatteryPowerLevel = 0.7 -- ratio; high end of max-kW slope
    M.minRegenFadeSpeed = jbeamData.minRegenFadeSpeed or 7
    M.maxRegenFadeSpeed = jbeamData.maxRegenFadeSpeed or 20
    M.maxEngineGeneratorTorque = jbeamData.maxEngineGeneratorTorque or 180
    M.maxFastChargePower = jbeamData.maxFastChargePower or 120
    M.autoStopBrakeThreshold = jbeamData.autoStopBrakeThreshold or 0
    M.autoStopBrakeTime = jbeamData.autoStopBrakeTime or 0
    M.engineMaxGeneratorRpm = jbeamData.maxGeneratorRpm or 0
    M.engineGeneratorMaxMinThrottle = jbeamData.engineGeneratorMaxMinThrottle or 0.15
    M.minThrottleAggressivenessRamp = jbeamData.minThrottleAggressivenessRamp or 2
    M.maxThrottleAggressivenessRamp = jbeamData.maxThrottleAggressivenessRamp or 15
    M.minThrottleAggressivenessFactor = jbeamData.minThrottleAggressivenessFactor or 1
    M.maxThrottleAggressivenessFactor = jbeamData.maxThrottleAggressivenessFactor or 12
    M.maxRegenCoef = jbeamData.maxRegenCoef or 0.8
    M.regenEfficiency = jbeamData.regenEfficiency or 0.65
    M.regenPercentOfBrakingForce = jbeamData.regenPercentOfBrakingForce or 0.25
    M.evPercentOfTotalPower = jbeamData.evPercentOfTotalPower or 0.65
    M.idlePowerDraw = jbeamData.idlePowerDraw or 1 -- kW
    M.evCreepThrottle = jbeamData.evCreepThrottle or 0.18
    M.evCoastThrottle = jbeamData.evCoastThrottle or -0.05
    M.chargeThrottleIFactor = jbeamData.chargeThrottleIFactor or 2.75
    M.startEvMode = jbeamData.startEvMode == true
    M.forceEvMode = jbeamData.forceEvMode == true

    torqueDemandSmoother =
        newTemporalSmoothingNonLinear(constants.torqueDemandSmoothingIn, constants.torqueDemandSmoothingOut)
end

local function reset()
    timeSinceInit = 0
    state = "init"
    nextState = nil
    
    init(M.jbeamData)
    M.initSecondStage()
end

local function initSounds()
    sounds.inverterWhine =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/inverter.ogg",
        "AudioDefaultLoop3D",
        "inverterWhine",
        engine.engineNodeID
    )

    sounds.motorRamp =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/motor_ramp.ogg",
        "AudioDefaultLoop3D",
        "motorRamp",
        engine.engineNodeID
    )
end

local function initSecondStage()
    engine = powertrain.getDevice("mainEngine")
    gearbox = powertrain.getDevice("gearbox")
    battery = energyStorage.getStorage("tractionBattery")

    if not gearbox or gearbox.type ~= "ecvtGearbox" then
        gearbox = nil
    end

    if battery then
        M.minBatteryPowerLevel = battery.jbeamData.minWorkingSoc or M.minBatteryPowerLevel
        M.maxBatteryPowerLevel = battery.jbeamData.maxWorkingSoc or M.maxBatteryPowerLevel
    end

    if engine then
        if not M.jbeamData.maxGeneratorRpm then
            M.engineMaxGeneratorRpm = engine:getTorqueData().maxPowerRPM
        end

        if not M.jbeamData.maxFastChargePower then
            -- Calculate a better default for this

            M.maxFastChargePower =
                M.engineMaxGeneratorRpm * constants.hpToKw * (M.maxEngineGeneratorTorque / constants.torqueToHpCoef)
        end
    end

    for i = 0, wheels.wheelCount - 1, 1 do
        initialWheelBrakeSplit[i] = {wheels.wheels[i].brakeInputSplit, wheels.wheels[i].brakeSplitCoef}
    end

    if M.startEvMode then
        electrics.values.evMode = 1
    end
end

M.updateGFX = updateGFX
M.init = init
M.initSounds = initSounds
M.initSecondStage = initSecondStage
M.reset = reset

return M
